/*
 *  <Short Description>
 *  Copyright (C) 2024  Brett Terpstra
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <functions.h>
#include <random>
#include <blt/gfx/stb/stb_perlin.h>

inline float protect_div(float x, float y, float d = 0)
{
    if (y == 0)
        return d;
    return x / y;
}

void f_x(image& img, float x, float y, blt::size_t argc, const image** argv, const data_t& extra_data)
{
    img = x;
}

void f_y(image& img, float x, float y, blt::size_t argc, const image** argv, const data_t& extra_data)
{
    img = y;
}

void f_random(image& img, float x, float y, blt::size_t argc, const image** argv, const data_t& extra_data)
{
    static std::random_device dev;
    static std::mt19937_64 engine(dev());
    static std::uniform_real_distribution<float> dist(0, 1);
    img = dist(engine);
}

void f_noise(image& img, float x, float y, blt::size_t argc, const image** argv, const data_t& extra_data)
{
    float z = 0.5;
    
    float scaleX = width;
    float scaleY = height;
    if (argc >= 1)
        scaleX *= argv[0].get().x();
    if (argc > 1)
        scaleY *= argv[1].get().x();
    
    auto val = stb_perlin_noise3(protect_div(x, scaleX, z),
                                 protect_div(y, scaleY, z),
                                 z, 0, 0, 0);
    img.set({val, val, val}, x, y);
}

void f_cnoise(image& img, float x, float y, blt::size_t argc, const image** argv, const data_t& extra_data)
{
    float z = 0.5;
    
    float scaleRX = width;
    float scaleGX = width;
    float scaleBX = width;
    float scaleRY = height;
    float scaleGY = height;
    float scaleBY = height;
    if (argc >= 1)
        scaleRX *= argv[0].get().x();
    if (argc > 1)
        scaleRY *= argv[1].get().x();
    if (argc >= 3)
        scaleRX *= argv[2].get().x();
    if (argc > 3)
        scaleRY *= argv[3].get().x();
    if (argc >= 5)
        scaleRX *= argv[4].get().x();
    if (argc > 5)
        scaleRY *= argv[5].get().x();
    
    auto valR = stb_perlin_noise3(protect_div(x, scaleRX, z),
                                 protect_div(y, scaleRY, z),
                                 z, 0, 0, 0);
    auto valG = stb_perlin_noise3(protect_div(x, scaleGX, z),
                                 protect_div(y, scaleGY, z),
                                 z, 0, 0, 0);
    auto valB = stb_perlin_noise3(protect_div(x, scaleBX, z),
                                 protect_div(y, scaleBY, z),
                                 z, 0, 0, 0);
    img.set({valR, valG, valB}, x, y);
}

void f_scalar(image& img, float x, float y, blt::size_t argc, const image** argv, const data_t& extra_data)
{
    img = extra_data[0];
}

void f_color(image& img, float x, float y, blt::size_t argc, const image** argv, const data_t& extra_data)
{
    img = blt::vec3{extra_data[0], extra_data[1], extra_data[2]};
}

void f_log(image& img, float x, float y, blt::size_t argc, const image** argv, const data_t& extra_data)
{

}

void f_sqrt(image& img, float x, float y, blt::size_t argc, const image** argv, const data_t& extra_data)
{

}

void f_sin(image& img, float x, float y, blt::size_t argc, const image** argv, const data_t& extra_data)
{

}

void f_cos(image& img, float x, float y, blt::size_t argc, const image** argv, const data_t& extra_data)
{

}

void f_atan(image& img, float x, float y, blt::size_t argc, const image** argv, const data_t& extra_data)
{

}

void f_add(image& img, float x, float y, blt::size_t argc, const image** argv, const data_t& extra_data)
{

}

void f_sub(image& img, float x, float y, blt::size_t argc, const image** argv, const data_t& extra_data)
{

}

void f_mul(image& img, float x, float y, blt::size_t argc, const image** argv, const data_t& extra_data)
{

}

void f_div(image& img, float x, float y, blt::size_t argc, const image** argv, const data_t& extra_data)
{

}

void f_or(image& img, float x, float y, blt::size_t argc, const image** argv, const data_t& extra_data)
{

}

void f_and(image& img, float x, float y, blt::size_t argc, const image** argv, const data_t& extra_data)
{

}

void f_xor(image& img, float x, float y, blt::size_t argc, const image** argv, const data_t& extra_data)
{

}

void f_if(image& img, float x, float y, blt::size_t argc, const image** argv, const data_t& extra_data)
{

}
